#include <casa/BasicMath/Math.h>
#include <tables/Tables.h>
#include <casa/Arrays.h>
#include <tables/Tables/TableIter.h>
//#include <casa/Quanta/MVTime.h>

#include <iostream>

#include "MS_File.h"

using namespace WSRT;
using namespace casa;

//===============>>>  MS_File::MS_File  <<<===============

MS_File::MS_File(const std::string& msname)
{
  MSName = msname;
  MS     = new MeasurementSet(MSName, Table::Update);
  init();
}

//===============>>>  MS_File::~MS_File  <<<===============

MS_File::~MS_File()
{
  delete MS;
}

//===============>>> MS_File::init  <<<===============

void MS_File::init()
{
  //Number of samples
  itsNumSamples                    = (*MS).nrow();    
  
  //Number of Fields
  MSField fields                   = (*MS).field();
  itsNumFields                     = fields.nrow();
  
  //Number of Antennae    
  MSAntenna antennae               = (*MS).antenna();
  itsNumAntennae                   = antennae.nrow();
  
  //Antenna Names
  ROScalarColumn<String>           ANT_NAME_col(antennae, "NAME");
  Vector<String>         ant_names = ANT_NAME_col.getColumn();
  ant_names.tovector(itsAntennaNames);
  
  //Number of channels in the Band
  MSSpectralWindow spectral_window = (*MS).spectralWindow();
  ROScalarColumn<Int>              NUM_CHAN_col(spectral_window, "NUM_CHAN");
  itsNumChannels                   = NUM_CHAN_col(0);
  
  //Number of polarizations
  MSPolarization      polarization = (*MS).polarization();
  ROScalarColumn<Int>              NUM_CORR_col(polarization, "NUM_CORR");
  itsNumPolarizations              = NUM_CORR_col(0);
  ROArrayColumn<Int>               CORR_TYPE_col(polarization, "CORR_TYPE");
  itsPolarizations.resize(itsNumPolarizations);
  CORR_TYPE_col.get(0, itsPolarizations);
  
  //calculate theoretical noise level
  ROScalarColumn<Double>           EXPOSURE_col((*MS), "EXPOSURE");
  Double exposure                  = EXPOSURE_col(0);

  ROScalarColumn<Double>           TOTAL_BANDWIDTH_col(spectral_window, "TOTAL_BANDWIDTH");
  Double bandwidth                 = TOTAL_BANDWIDTH_col(0) / itsNumChannels;
  
  itsNoiseLevel                    = 1.0 / sqrt(bandwidth * exposure);
    
  //calculate number of timeslots
  ROScalarColumn<Double>           INTERVAL_col((*MS), "INTERVAL");
  Double interval                  = INTERVAL_col(0);
  
  //Number of timeslots
  ROScalarColumn<Double>            TIME_CENTROID_col((*MS), "TIME_CENTROID");
  Double firstdate                 = TIME_CENTROID_col(0);
  Double lastdate                  = TIME_CENTROID_col(itsNumSamples-1);
  
  itsNumTimeslots                  = (int)((lastdate-firstdate)/interval) + 1;
  
  //calculate number of baselines.
  itsNumPairs = (itsNumAntennae) * (itsNumAntennae + 1) / 2; //Triangular numbers formula

  //calculate number of Bands
  itsNumBands                      = itsNumSamples / (itsNumPairs * itsNumTimeslots);
}

//===============>>> MS_File::BaselineIterator  <<<===============

MSAntenna MS_File::antenna()
{  
  return (*MS).antenna();
}

//===============>>> MS_File::BaselineIterator  <<<===============

TableIterator MS_File::TimeslotIterator()
{  
  Block<String> ms_iteration_variables(1);
  ms_iteration_variables[0] = "TIME_CENTROID";
  
  return TableIterator((*MS), ms_iteration_variables);
}  

//===============>>> MS_File::BaselineIterator  <<<===============

TableIterator MS_File::TimeAntennaIterator()
{  
  Block<String> ms_iteration_variables(4);
  ms_iteration_variables[0] = "TIME_CENTROID";
  ms_iteration_variables[1] = "DATA_DESC_ID";
  ms_iteration_variables[2] = "ANTENNA1";
  ms_iteration_variables[3] = "ANTENNA2";
  
  return TableIterator((*MS), ms_iteration_variables);
}  

//===============>>> MS_File::WriteDataPointFlags  <<<===============

void   MS_File::WriteDataPointFlags(TableIterator* flag_iter, 
                                    Matrix<Bool>* Flags,
                                    Bool FlagCompleteRow,
                                    bool ExistingFlags)
{
  Table         FlagTable = (*flag_iter).table();
//  ROTableVector<Int>      antenna1(FlagTable, "ANTENNA1");
//  ROTableVector<Int>      antenna2(FlagTable, "ANTENNA2");
//  ROTableVector<Int>      bandnr  (FlagTable, "DATA_DESC_ID");
//  ROTableVector<Double>   time    (FlagTable, "TIME_CENTROID");
  ArrayColumn  <Bool>     flags   (FlagTable, "FLAG");
  ScalarColumn <Bool>     flag_row(FlagTable, "FLAG_ROW");
//  ArrayColumn<String> flag_desc((*TimeslotTable), "FLAG_CATEGORY");
  if (ExistingFlags)
  {
    Matrix<Bool> OldFlags;
    Bool         OldFlagCompleteRow;
    flags.get(0, OldFlags);
    flag_row.get(0, OldFlagCompleteRow);
    flags.put(0, OldFlags || (*Flags));
  }
  else
  {
    flags.put(0, (*Flags));
  }
  flag_row.put(0, FlagCompleteRow);
}
//===============>>> MS_File  <<<===============

